"""Determine notebook or cell language"""

# Jupyter magic commands that are also languages
_JUPYTER_LANGUAGES = ['R', 'bash', 'sh', 'python', 'python2', 'python3', 'javascript', 'js', 'perl',
                      'html', 'latex', 'markdown', 'pypy', 'ruby', 'script', 'svg',
                      'matlab', 'octave', 'idl', 'robotframework', 'spark', 'sql']

# Supported file extensions (and languages)
# Please add more languages here (and add a few tests) - see CONTRIBUTING.md
_SCRIPT_EXTENSIONS = {'.py': {'language': 'python', 'comment': '#'},
                      '.R': {'language': 'R', 'comment': '#'},
                      '.r': {'language': 'R', 'comment': '#'},
                      '.jl': {'language': 'julia', 'comment': '#'},
                      '.cpp': {'language': 'c++', 'comment': '//'},
                      '.ss': {'language': 'scheme', 'comment': ';;'},
                      '.clj': {'language': 'clojure', 'comment': ';;'},
                      '.scm': {'language': 'scheme', 'comment': ';;'},
                      '.sh': {'language': 'bash', 'comment': '#'},
                      '.ps1': {'language': 'powershell', 'comment': '#'},
                      '.q': {'language': 'q', 'comment': '/'},
                      '.m': {'language': 'matlab', 'comment': '%'},
                      '.pro': {'language': 'idl', 'comment': ';'},
                      '.js': {'language': 'javascript', 'comment': '//'},
                      '.ts': {'language': 'typescript', 'comment': '//'},
                      '.scala': {'language': 'scala', 'comment': '//'},
                      '.rs': {'language': 'rust', 'comment': '//'},
                      '.robot': {'language': 'robotframework', 'comment': '#'}}

_COMMENT_CHARS = [_SCRIPT_EXTENSIONS[ext]['comment'] for ext in _SCRIPT_EXTENSIONS if
                  _SCRIPT_EXTENSIONS[ext]['comment'] != '#']

_COMMENT = {_SCRIPT_EXTENSIONS[ext]['language']: _SCRIPT_EXTENSIONS[ext]['comment'] for ext in _SCRIPT_EXTENSIONS}
_JUPYTER_LANGUAGES = set(_JUPYTER_LANGUAGES).union(_COMMENT.keys())


def default_language_from_metadata_and_ext(metadata, ext):
    """Return the default language given the notebook metadata, and a file extension"""
    default_from_ext = _SCRIPT_EXTENSIONS.get(ext, {}).get('language')

    language = (metadata.get('jupytext', {}).get('main_language')
                or metadata.get('kernelspec', {}).get('language')
                or default_from_ext)

    if language is None or language == 'R':
        return language

    if language.startswith('C++'):
        return 'c++'

    return language.lower()


def same_language(kernel_language, language):
    """Are those the same language?"""
    if kernel_language == language:
        return True
    if kernel_language.lower() == language:
        return True
    if kernel_language.startswith('C++') and language == 'c++':
        return True
    if kernel_language == 'octave' and language == 'matlab':
        return True
    return False


def set_main_and_cell_language(metadata, cells, ext):
    """Set main language for the given collection of cells, and
    use magics for cells that use other languages"""
    main_language = default_language_from_metadata_and_ext(metadata, ext)

    if main_language is None:
        languages = {'python': 0.5}
        for cell in cells:
            if 'language' in cell['metadata']:
                language = cell['metadata']['language']
                languages[language] = languages.get(language, 0.0) + 1

        main_language = max(languages, key=languages.get)

    # save main language when no kernel is set
    if 'language' not in metadata.get('kernelspec', {}) and cells:
        metadata.setdefault('jupytext', {})['main_language'] = main_language

    # Remove 'language' meta data and add a magic if not main language
    for cell in cells:
        if 'language' in cell['metadata']:
            language = cell['metadata'].pop('language')
            if language != main_language and language in _JUPYTER_LANGUAGES:
                if 'magic_args' in cell['metadata']:
                    magic_args = cell['metadata'].pop('magic_args')
                    cell['source'] = u'%%{} {}\n'.format(language, magic_args) + cell['source']
                else:
                    cell['source'] = u'%%{}\n'.format(language) + cell['source']


def cell_language(source):
    """Return cell language and language options, if any"""
    if source:
        line = source[0]
        if line.startswith('%%'):
            magic = line[2:]
            if ' ' in magic:
                lang, magic_args = magic.split(' ', 1)
            else:
                lang = magic
                magic_args = ''

            if lang in _JUPYTER_LANGUAGES:
                source.pop(0)
                return lang, magic_args

    return None, None


def comment_lines(lines, prefix):
    """Return commented lines"""
    if not prefix:
        return lines
    return [prefix + ' ' + line if line else prefix for line in lines]



# ---
# jupyter:
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Part one - various cells

# %% [markdown]
# Here we have a markdown cell
#
#
# with two blank lines

# %% [markdown]
# Now we have a markdown cell
# with a code block inside it
#
# ```python
# 1 + 1
# ```
#
# After that cell we'll have a code cell

# %%
2 + 2


3 + 3

# %% [markdown]
# Followed by a raw cell

# %% [raw]
# This is 
# the content
# of the raw cell

# %% [markdown]
# # Part two - cell metadata

# %% [markdown] key="value"
# This is a markdown cell with cell metadata `{"key": "value"}`

# %% .class tags=["parameters"]
"""This is a code cell with metadata `{"tags":["parameters"], ".class":null}`"""

# %% [raw] key="value"
# This is a raw cell with cell metadata `{"key": "value"}`



import os
import pytest
from shutil import copyfile
from jupytext.compare import compare
from nbformat.v4.nbbase import new_notebook, new_code_cell
from .utils import list_notebooks, requires_black, requires_flake8, requires_autopep8

from jupytext import read, write
from jupytext.cli import system, jupytext, pipe_notebook
from jupytext.combine import black_invariant
from jupytext.header import _DEFAULT_NOTEBOOK_METADATA


@requires_black
@pytest.mark.parametrize('nb_file', list_notebooks('ipynb_py'))
def test_apply_black_on_python_notebooks(nb_file, tmpdir):
    tmp_ipynb = str(tmpdir.join('notebook.ipynb'))
    tmp_py = str(tmpdir.join('notebook.py'))
    copyfile(nb_file, tmp_ipynb)

    jupytext(args=[tmp_ipynb, '--to', 'py:percent'])
    system('black', tmp_py)
    jupytext(args=[tmp_py, '--to', 'ipynb', '--update'])

    nb1 = read(nb_file)
    nb2 = read(tmp_ipynb)
    nb3 = read(tmp_py)

    assert len(nb1.cells) == len(nb2.cells)
    assert len(nb1.cells) == len(nb3.cells)
    for c1, c2 in zip(nb1.cells, nb2.cells):
        # same content (almost)
        assert black_invariant(c1.source) == black_invariant(c2.source)
        # python representation is pep8
        assert 'lines_to_next_cell' not in c2.metadata
        # outputs are preserved
        assert c1.cell_type == c2.cell_type
        if c1.cell_type == 'code':
            compare(c1.outputs, c2.outputs)

    compare(nb1.metadata, nb2.metadata)


def test_black_invariant():
    text_org = """long_string = "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \\
              "bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb"
"""
    text_black = """long_string = (
    "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
    "bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb"
)
"""
    assert black_invariant(text_org) == black_invariant(text_black)


@requires_black
def test_pipe_into_black():
    nb_org = new_notebook(cells=[new_code_cell('1        +1')])
    nb_dest = new_notebook(cells=[new_code_cell('1 + 1')])

    nb_pipe = pipe_notebook(nb_org, 'black')
    compare(nb_pipe, nb_dest)


@requires_autopep8
def test_pipe_into_autopep8():
    nb_org = new_notebook(cells=[new_code_cell('1        +1')])
    nb_dest = new_notebook(cells=[new_code_cell('1 + 1')])

    nb_pipe = pipe_notebook(nb_org, 'autopep8 -')
    compare(nb_pipe, nb_dest)


@requires_flake8
def test_pipe_into_flake8():
    # Notebook OK
    nb = new_notebook(cells=[new_code_cell('# correct code\n1 + 1')])
    pipe_notebook(nb, 'flake8', update=False)

    # Notebook not OK
    nb = new_notebook(cells=[new_code_cell('incorrect code')])
    with pytest.raises(SystemExit):
        pipe_notebook(nb, 'flake8', update=False)


@requires_black
@pytest.mark.parametrize('nb_file', list_notebooks('ipynb_py')[:1])
def test_apply_black_through_jupytext(tmpdir, nb_file):
    # Load real notebook metadata to get the 'auto' extension in --pipe-fmt to work
    metadata = read(nb_file).metadata

    nb_org = new_notebook(cells=[new_code_cell('1        +1')], metadata=metadata)
    nb_black = new_notebook(cells=[new_code_cell('1 + 1')], metadata=metadata)

    os.makedirs(str(tmpdir.join('notebook_folder')))
    os.makedirs(str(tmpdir.join('script_folder')))

    tmp_ipynb = str(tmpdir.join('notebook_folder').join('notebook.ipynb'))
    tmp_py = str(tmpdir.join('script_folder').join('notebook.py'))

    # Black in place
    write(nb_org, tmp_ipynb)
    jupytext([tmp_ipynb, '--pipe', 'black'])
    nb_now = read(tmp_ipynb)
    compare(nb_now, nb_black)

    # Write to another folder using dots
    script_fmt = os.path.join('..', 'script_folder//py:percent')
    write(nb_org, tmp_ipynb)
    jupytext([tmp_ipynb, '--to', script_fmt, '--pipe', 'black'])
    assert os.path.isfile(tmp_py)
    nb_now = read(tmp_py)
    nb_now.metadata = metadata
    compare(nb_now, nb_black)
    os.remove(tmp_py)

    # Map to another folder based on file name
    write(nb_org, tmp_ipynb)
    jupytext([tmp_ipynb, '--from', 'notebook_folder//ipynb', '--to', 'script_folder//py:percent',
              '--pipe', 'black', '--check', 'flake8'])
    assert os.path.isfile(tmp_py)
    nb_now = read(tmp_py)
    nb_now.metadata = metadata
    compare(nb_now, nb_black)


@requires_black
@pytest.mark.parametrize('nb_file', list_notebooks('ipynb_py'))
def test_apply_black_and_sync_on_paired_notebook(tmpdir, nb_file):
    # Load real notebook metadata to get the 'auto' extension in --pipe-fmt to work
    metadata = read(nb_file).metadata
    metadata['jupytext'] = {'formats': 'ipynb,py'}
    assert 'language_info' in metadata

    nb_org = new_notebook(cells=[new_code_cell('1        +1')], metadata=metadata)
    nb_black = new_notebook(cells=[new_code_cell('1 + 1')], metadata=metadata)

    tmp_ipynb = str(tmpdir.join('notebook.ipynb'))
    tmp_py = str(tmpdir.join('notebook.py'))

    # Black in place
    write(nb_org, tmp_ipynb)
    jupytext([tmp_ipynb, '--pipe', 'black', '--sync'])

    nb_now = read(tmp_ipynb)
    compare(nb_now, nb_black)
    assert 'language_info' in nb_now.metadata

    nb_now = read(tmp_py)
    nb_now.metadata['jupytext'].pop('text_representation')
    nb_black.metadata = {key: nb_black.metadata[key] for key in nb_black.metadata
                         if key in _DEFAULT_NOTEBOOK_METADATA}
    compare(nb_now, nb_black)


@requires_black
def test_apply_black_on_markdown_notebook(tmpdir):
    text = """---
jupyter:
  kernelspec:
    display_name: Python 3
    language: python
    name: python3
  language_info:
    codemirror_mode:
      name: ipython
      version: 3
    file_extension: .py
    mimetype: text/x-python
    name: python
    nbconvert_exporter: python
    pygments_lexer: ipython3
    version: 3.7.4
---

```python
1    +     \
2+3\
+4
```
"""
    tmp_md = str(tmpdir.join('test.md'))
    with open(tmp_md, 'w') as fp:
        fp.write(text)

    jupytext([tmp_md, '--pipe', 'black'])

    nb = read(tmp_md)
    compare(nb.cells, [new_code_cell('1 + 2 + 3 + 4')])


@requires_black
def test_black_through_tempfile(
        tmpdir,
        text="""```python
1 +    2 \
+ 3
```
""",
        black="""```python
1 + 2 + 3
```
"""):
    tmp_md = str(tmpdir.join('notebook.md'))
    with open(tmp_md, 'w') as fp:
        fp.write(text)
    jupytext([tmp_md, '--pipe', 'black {}'])
    with open(tmp_md) as fp:
        compare(fp.read(), black)



"""Combine source and outputs from two notebooks
"""
import re
from copy import copy
from .cell_metadata import _IGNORE_CELL_METADATA, _JUPYTEXT_CELL_METADATA
from .header import _DEFAULT_NOTEBOOK_METADATA
from .metadata_filter import filter_metadata
from .formats import long_form_one_format

_BLANK_LINE = re.compile(r'^\s*$')


def black_invariant(text, chars=None):
    """Remove characters that may be changed when reformatting the text with black"""
    if chars is None:
        chars = [' ', '\t', '\n', ',', "'", '"', '(', ')', '\\']

    for char in chars:
        text = text.replace(char, '')
    return text


def same_content(ref, test):
    """Is the content of two cells the same, up to reformating by black"""
    return black_invariant(ref) == black_invariant(test)


def combine_inputs_with_outputs(nb_source, nb_outputs, fmt=None):
    """Copy outputs of the second notebook into
    the first one, for cells that have matching inputs"""

    output_code_cells = [cell for cell in nb_outputs.cells if cell.cell_type == 'code']
    output_other_cells = [cell for cell in nb_outputs.cells if cell.cell_type != 'code']

    fmt = long_form_one_format(fmt)
    text_repr = nb_source.metadata.get('jupytext', {}).get('text_representation', {})
    ext = fmt.get('extension') or text_repr.get('extension')
    format_name = fmt.get('format_name') or text_repr.get('format_name')

    nb_outputs_filtered_metadata = copy(nb_outputs.metadata)
    filter_metadata(nb_outputs_filtered_metadata,
                    nb_source.metadata.get('jupytext', {}).get('notebook_metadata_filter'),
                    _DEFAULT_NOTEBOOK_METADATA)

    for key in nb_outputs.metadata:
        if key not in nb_outputs_filtered_metadata:
            nb_source.metadata[key] = nb_outputs.metadata[key]

    source_is_md_version_one = ext in ['.md', '.markdown', '.Rmd'] and text_repr.get('format_version') == '1.0'
    if nb_source.metadata.get('jupytext', {}).get('formats') or ext in ['.md', '.markdown', '.Rmd']:
        nb_source.metadata.get('jupytext', {}).pop('text_representation', None)

    if not nb_source.metadata.get('jupytext', {}):
        nb_source.metadata.pop('jupytext', {})

    for cell in nb_source.cells:
        # Remove outputs to warranty that trust of returned notebook is that of second notebook
        if cell.cell_type == 'code':
            cell.execution_count = None
            cell.outputs = []

            for i, ocell in enumerate(output_code_cells):
                if same_content(cell.source, ocell.source):
                    # Fill outputs with that of second notebook
                    cell.execution_count = ocell.execution_count
                    cell.outputs = ocell.outputs

                    # Append cell metadata that was filtered
                    if format_name in ['bare', 'sphinx'] or source_is_md_version_one:
                        ocell_filtered_metadata = {}
                    else:
                        ocell_filtered_metadata = copy(ocell.metadata)
                        filter_metadata(ocell_filtered_metadata,
                                        nb_source.metadata.get('jupytext', {}).get('cell_metadata_filter'),
                                        _IGNORE_CELL_METADATA)

                    for key in ocell.metadata:
                        if key not in ocell_filtered_metadata and key not in _JUPYTEXT_CELL_METADATA:
                            cell.metadata[key] = ocell.metadata[key]

                    output_code_cells = output_code_cells[(i + 1):]
                    break
        else:
            for i, ocell in enumerate(output_other_cells):
                if cell.cell_type == ocell.cell_type and same_content(cell.source, ocell.source):
                    if format_name in ['spin', 'bare', 'sphinx'] or source_is_md_version_one:
                        ocell_filtered_metadata = {}
                    else:
                        ocell_filtered_metadata = copy(ocell.metadata)
                        filter_metadata(ocell_filtered_metadata,
                                        nb_source.metadata.get('jupytext', {}).get('cell_metadata_filter'),
                                        _IGNORE_CELL_METADATA)

                    for key in ocell.metadata:
                        if key not in ocell_filtered_metadata:
                            cell.metadata[key] = ocell.metadata[key]

                    output_other_cells = output_other_cells[(i + 1):]
                    break

    return nb_source



"""Read notebook cells from their text representation"""

import re
from copy import copy
from nbformat.v4.nbbase import new_code_cell, new_raw_cell, new_markdown_cell
from .languages import _SCRIPT_EXTENSIONS

# Sphinx Gallery is an optional dependency. And we intercept the SyntaxError for #301
try:
    from sphinx_gallery.notebook import rst2md
except (ImportError, SyntaxError):
    rst2md = None

from .cell_metadata import is_active, text_to_metadata, is_json_metadata, rmd_options_to_metadata
from .languages import _JUPYTER_LANGUAGES
from .stringparser import StringParser
from .magics import uncomment_magic, is_magic, unescape_code_start, need_explicit_marker
from .pep8 import pep8_lines_between_cells

_BLANK_LINE = re.compile(r"^\s*$")
_PY_INDENTED = re.compile(r"^\s")


def uncomment(lines, prefix='#'):
    """Remove prefix and space, or only prefix, when possible"""
    if not prefix:
        return lines
    prefix_and_space = prefix + ' '
    length_prefix = len(prefix)
    length_prefix_and_space = len(prefix_and_space)
    return [line[length_prefix_and_space:] if line.startswith(prefix_and_space)
            else (line[length_prefix:] if line.startswith(prefix) else line)
            for line in lines]


def paragraph_is_fully_commented(lines, comment, main_language):
    """Is the paragraph fully commented?"""
    for i, line in enumerate(lines):
        if line.startswith(comment):
            if line[len(comment):].lstrip().startswith(comment):
                continue
            if is_magic(line, main_language):
                return False
            continue
        return i > 0 and _BLANK_LINE.match(line)
    return True


def next_code_is_indented(lines):
    """Is the next unescaped line indented?"""
    for line in lines:
        if _BLANK_LINE.match(line):
            continue
        return _PY_INDENTED.match(line)
    return False


def count_lines_to_next_cell(cell_end_marker, next_cell_start, total, explicit_eoc):
    """How many blank lines between end of cell marker and next cell?"""
    if cell_end_marker < total:
        lines_to_next_cell = next_cell_start - cell_end_marker
        if explicit_eoc:
            lines_to_next_cell -= 1
        if next_cell_start >= total:
            lines_to_next_cell += 1
        return lines_to_next_cell

    return 1


def last_two_lines_blank(source):
    """Are the two last lines blank, and not the third last one?"""
    if len(source) < 3:
        return False
    return not _BLANK_LINE.match(source[-3]) and _BLANK_LINE.match(source[-2]) and _BLANK_LINE.match(source[-1])


class BaseCellReader(object):
    """A class that can read notebook cells from their text representation"""
    default_comment_magics = None
    lines_to_next_cell = 1

    start_code_re = None
    simple_start_code_re = None
    end_code_re = None

    # How to make code inactive
    comment = ''

    # Any specific prefix for lines in markdown cells (like in R spin format?)
    markdown_prefix = None

    def __init__(self, fmt=None, default_language=None):
        """Create a cell reader with empty content"""
        if not fmt:
            fmt = {}
        self.ext = fmt.get('extension')
        self.default_language = default_language or _SCRIPT_EXTENSIONS.get(self.ext, {}).get('language', 'python')
        self.comment_magics = fmt.get('comment_magics', self.default_comment_magics)
        self.use_runtools = fmt.get('use_runtools', False)
        self.format_version = fmt.get('format_version')
        self.metadata = None
        self.org_content = []
        self.content = []
        self.explicit_soc = None
        self.explicit_eoc = None
        self.cell_type = None
        self.language = None
        self.cell_metadata_json = fmt.get('cell_metadata_json', False)

    def read(self, lines):
        """Read one cell from the given lines, and return the cell,
        plus the position of the next cell
        """

        # Do we have an explicit code marker on the first line?
        self.metadata_and_language_from_option_line(lines[0])

        if self.metadata and 'language' in self.metadata:
            self.language = self.metadata.pop('language')

        # Parse cell till its end and set content, lines_to_next_cell
        pos_next_cell = self.find_cell_content(lines)

        if self.cell_type == 'code':
            new_cell = new_code_cell
        elif self.cell_type == 'markdown':
            new_cell = new_markdown_cell
        else:
            new_cell = new_raw_cell

        if not self.metadata:
            self.metadata = {}

        if self.ext == '.py':
            expected_blank_lines = pep8_lines_between_cells(self.org_content or [''], lines[pos_next_cell:], self.ext)
        else:
            expected_blank_lines = 1

        if self.lines_to_next_cell != expected_blank_lines:
            self.metadata['lines_to_next_cell'] = self.lines_to_next_cell

        if self.language:
            self.metadata['language'] = self.language

        return new_cell(source='\n'.join(self.content), metadata=self.metadata), pos_next_cell

    def metadata_and_language_from_option_line(self, line):
        """Parse code options on the given line. When a start of a code cell
        is found, self.metadata is set to a dictionary."""
        if self.start_code_re.match(line):
            self.language, self.metadata = self.options_to_metadata(self.start_code_re.findall(line)[0])

    def options_to_metadata(self, options):
        """Return language (str) and metadata (dict) from the option line"""
        raise NotImplementedError("Option parsing must be implemented in a sub class")

    def find_cell_end(self, lines):
        """Return position of end of cell marker, and position
        of first line after cell"""
        raise NotImplementedError('This method must be implemented in a sub class')

    def find_cell_content(self, lines):
        """Parse cell till its end and set content, lines_to_next_cell.
        Return the position of next cell start"""
        cell_end_marker, next_cell_start, self.explicit_eoc = self.find_cell_end(lines)

        # Metadata to dict
        if self.metadata is None:
            cell_start = 0
            self.metadata = {}
        else:
            cell_start = 1

        # Cell content
        source = lines[cell_start:cell_end_marker]
        self.org_content = copy(source)

        # Exactly two empty lines at the end of cell (caused by PEP8)?
        if self.ext == '.py' and self.explicit_eoc:
            if last_two_lines_blank(source):
                source = source[:-2]
                lines_to_end_of_cell_marker = 2
            else:
                lines_to_end_of_cell_marker = 0

            pep8_lines = pep8_lines_between_cells(source, lines[cell_end_marker:], self.ext)
            if lines_to_end_of_cell_marker != (0 if pep8_lines == 1 else 2):
                self.metadata['lines_to_end_of_cell_marker'] = lines_to_end_of_cell_marker

        # Uncomment content
        self.explicit_soc = cell_start > 0
        self.content = self.extract_content(source)

        # Is this an inactive cell?
        if self.cell_type == 'code':
            if not is_active('.ipynb', self.metadata):
                if self.metadata.get('active') == '':
                    del self.metadata['active']
                self.cell_type = 'raw'
            elif self.ext in ['.md', '.markdown'] and not self.language:
                # Markdown files in version >= 1.3 represent code chunks with no language as Markdown cells
                if self.format_version not in ['1.0', '1.1']:
                    self.cell_type = 'markdown'
                    self.explicit_eoc = False
                    cell_end_marker += 1
                    self.content = lines[:cell_end_marker]
                # Previous versions mapped those to raw cells
                else:
                    self.cell_type = 'raw'

        # Explicit end of cell marker?
        if (next_cell_start + 1 < len(lines) and
                _BLANK_LINE.match(lines[next_cell_start]) and
                not _BLANK_LINE.match(lines[next_cell_start + 1])):
            next_cell_start += 1
        elif (self.explicit_eoc and next_cell_start + 2 < len(lines) and
              _BLANK_LINE.match(lines[next_cell_start]) and
              _BLANK_LINE.match(lines[next_cell_start + 1]) and
              not _BLANK_LINE.match(lines[next_cell_start + 2])):
            next_cell_start += 2

        self.lines_to_next_cell = count_lines_to_next_cell(
            cell_end_marker,
            next_cell_start,
            len(lines),
            self.explicit_eoc)

        return next_cell_start

    def uncomment_code_and_magics(self, lines):
        """Uncomment code and possibly commented magic commands"""
        raise NotImplementedError('This method must be implemented in a sub class')

    def extract_content(self, lines):
        # Code cells with just a multiline string become Markdown cells
        if self.ext == '.py' and not is_active(self.ext, self.metadata, self.cell_type == 'code'):
            content = '\n'.join(lines).strip()
            for triple_quote in ['"""', "'''"]:
                if content.startswith(triple_quote) and content.endswith(triple_quote) and len(content) >= 6:
                    left = right = triple_quote
                    content = content[3:-3]
                    # Trim first/last line return
                    if content.startswith('\n'):
                        content = content[1:]
                        left = triple_quote + '\n'
                    if content.endswith('\n'):
                        content = content[:-1]
                        right = '\n' + triple_quote
                    if len(left) == len(right) == 4:
                        self.metadata['cell_marker'] = left[:3]
                    else:
                        self.metadata['cell_marker'] = left + ',' + right
                    return content.splitlines()

        if not is_active(self.ext, self.metadata) or \
                ('active' not in self.metadata and self.language and self.language != self.default_language):
            return uncomment(lines, self.comment if self.ext not in ['.r', '.R'] else '#')

        return self.uncomment_code_and_magics(lines)


class MarkdownCellReader(BaseCellReader):
    """Read notebook cells from Markdown documents"""
    comment = ''
    start_code_re = re.compile(r"^```({})($|\s(.*)$)".format('|'.join(
        _JUPYTER_LANGUAGES.union({str.upper(lang) for lang in _JUPYTER_LANGUAGES})).replace('+', '\\+')))
    non_jupyter_code_re = re.compile(r"^```")
    end_code_re = re.compile(r"^```\s*$")
    start_region_re = re.compile(r"^<!--\s*#(region|markdown|md|raw)(.*)-->\s*$")
    end_region_re = None
    default_comment_magics = False

    def __init__(self, fmt=None, default_language=None):
        super(MarkdownCellReader, self).__init__(fmt, default_language)
        self.split_at_heading = (fmt or {}).get('split_at_heading', False)
        self.in_region = False
        self.in_raw = False
        if self.format_version in ['1.0', '1.1'] and self.ext != '.Rmd':
            # Restore the pattern used in Markdown <= 1.1
            self.start_code_re = re.compile(r"^```(.*)")
            self.non_jupyter_code_re = re.compile(r"^```\{")

    def metadata_and_language_from_option_line(self, line):
        match_region = self.start_region_re.match(line)
        if match_region:
            self.in_region = True
            groups = match_region.groups()
            region_name = groups[0]
            self.end_region_re = re.compile(r"^<!--\s*#end{}\s*-->\s*$".format(region_name))
            self.cell_metadata_json = self.cell_metadata_json or is_json_metadata(groups[1])
            title, self.metadata = text_to_metadata(groups[1], allow_title=True)
            if region_name == 'raw':
                self.cell_type = 'raw'
            else:
                self.cell_type = 'markdown'
            if title:
                self.metadata['title'] = title
            if region_name in ['markdown', 'md']:
                self.metadata['region_name'] = region_name
        elif self.start_code_re.match(line):
            self.language, self.metadata = self.options_to_metadata(self.start_code_re.findall(line)[0])
            # Cells with a .noeval attribute are markdown cells #347
            if self.metadata.get('.noeval', '') is None:
                self.cell_type = 'markdown'
                self.metadata = {}
                self.language = None

    def options_to_metadata(self, options):
        if isinstance(options, tuple):
            options = ' '.join(options)
        self.cell_metadata_json = self.cell_metadata_json or is_json_metadata(options)
        return text_to_metadata(options)

    def find_cell_end(self, lines):
        """Return position of end of cell marker, and position
        of first line after cell"""
        if self.in_region:
            for i, line in enumerate(lines):
                if self.end_region_re.match(line):
                    return i, i + 1, True
        elif self.metadata is None:
            # default markdown: (last) two consecutive blank lines, except when in code blocks
            self.cell_type = 'markdown'
            prev_blank = 0
            in_explicit_code_block = False
            in_indented_code_block = False

            for i, line in enumerate(lines):
                if in_explicit_code_block and self.end_code_re.match(line):
                    in_explicit_code_block = False
                    continue

                if prev_blank and line.startswith('    ') and not _BLANK_LINE.match(line):
                    in_indented_code_block = True
                    prev_blank = 0
                    continue

                if in_indented_code_block and not _BLANK_LINE.match(line) and not line.startswith('    '):
                    in_indented_code_block = False

                if in_indented_code_block or in_explicit_code_block:
                    continue

                if self.start_region_re.match(line):
                    if i > 1 and prev_blank:
                        return i - 1, i, False
                    return i, i, False

                if self.start_code_re.match(line):
                    # Cells with a .noeval attribute are markdown cells #347
                    _, metadata = self.options_to_metadata(self.start_code_re.findall(line)[0])
                    if metadata.get('.noeval', '') is None:
                        in_explicit_code_block = True
                        prev_blank = 0
                        continue

                    if i > 1 and prev_blank:
                        return i - 1, i, False
                    return i, i, False

                if self.non_jupyter_code_re.match(line):
                    if prev_blank >= 2:
                        return i - 2, i, True
                    in_explicit_code_block = True
                    prev_blank = 0
                    continue

                if self.split_at_heading and line.startswith('#') and prev_blank >= 1:
                    return i - 1, i, False

                if _BLANK_LINE.match(lines[i]):
                    prev_blank += 1
                elif prev_blank >= 2:
                    return i - 2, i, True
                else:
                    prev_blank = 0
        else:
            self.cell_type = 'code'
            for i, line in enumerate(lines):
                # skip cell header
                if i == 0:
                    continue
                if self.end_code_re.match(line):
                    return i, i + 1, True

        # End not found
        return len(lines), len(lines), False

    def uncomment_code_and_magics(self, lines):
        if self.cell_type == 'code' and self.comment_magics:
            lines = uncomment_magic(lines, self.language)
        return lines


class RMarkdownCellReader(MarkdownCellReader):
    """Read notebook cells from R Markdown notebooks"""
    comment = ''
    start_code_re = re.compile(r"^```{(.*)}\s*$")
    non_jupyter_code_re = re.compile(r"^```([^\{]|\s*$)")
    default_language = 'R'
    default_comment_magics = True

    def options_to_metadata(self, options):
        return rmd_options_to_metadata(options, self.use_runtools)

    def uncomment_code_and_magics(self, lines):
        if self.cell_type == 'code' and self.comment_magics and is_active(self.ext, self.metadata):
            uncomment_magic(lines, self.language or self.default_language)

        return lines


class ScriptCellReader(BaseCellReader):  # pylint: disable=W0223
    """Read notebook cells from scripts
    (common base for R and Python scripts)"""

    def uncomment_code_and_magics(self, lines):
        if self.cell_type == 'code' or self.comment != "#'":
            if self.comment_magics:
                if is_active(self.ext, self.metadata):
                    uncomment_magic(lines, self.language or self.default_language, explicitly_code=self.explicit_soc)
                    if self.cell_type == 'code' and not self.explicit_soc and \
                            need_explicit_marker(lines, self.language or self.default_language):
                        self.metadata['comment_questions'] = False
                else:
                    lines = uncomment(lines)

        if self.cell_type == 'code':
            return unescape_code_start(lines, self.ext, self.language or self.default_language)

        return uncomment(lines, self.markdown_prefix or self.comment)


class RScriptCellReader(ScriptCellReader):
    """Read notebook cells from R scripts written according
    to the knitr-spin syntax"""

    comment = "#'"
    markdown_prefix = "#'"
    default_language = 'R'
    start_code_re = re.compile(r"^#\+(.*)\s*$")
    default_comment_magics = True

    def options_to_metadata(self, options):
        return rmd_options_to_metadata('r ' + options, self.use_runtools)

    def find_cell_end(self, lines):
        """Return position of end of cell marker, and position
        of first line after cell"""
        if self.metadata is None and lines[0].startswith("#'"):
            self.cell_type = 'markdown'
            for i, line in enumerate(lines):
                if not line.startswith("#'"):
                    if _BLANK_LINE.match(line):
                        return i, i + 1, False
                    return i, i, False

            return len(lines), len(lines), False

        if self.metadata and 'cell_type' in self.metadata:
            self.cell_type = self.metadata.pop('cell_type')
        else:
            self.cell_type = 'code'

        parser = StringParser(self.language or self.default_language)
        for i, line in enumerate(lines):
            # skip cell header
            if self.metadata is not None and i == 0:
                continue

            if parser.is_quoted():
                parser.read_line(line)
                continue

            parser.read_line(line)

            if self.start_code_re.match(line) or (self.markdown_prefix and line.startswith(self.markdown_prefix)):
                if i > 0 and _BLANK_LINE.match(lines[i - 1]):
                    if i > 1 and _BLANK_LINE.match(lines[i - 2]):
                        return i - 2, i, False
                    return i - 1, i, False
                return i, i, False

            if _BLANK_LINE.match(line):
                if not next_code_is_indented(lines[i:]):
                    if i > 0:
                        return i, i + 1, False
                    if len(lines) > 1 and not _BLANK_LINE.match(lines[1]):
                        return 1, 1, False
                    return 1, 2, False

        return len(lines), len(lines), False


class LightScriptCellReader(ScriptCellReader):
    """Read notebook cells from plain Python or Julia files. Cells
    are identified by line breaks, unless they start with an
    explicit marker '# +' """
    default_comment_magics = True
    cell_marker_start = None
    cell_marker_end = None

    def __init__(self, fmt=None, default_language=None):
        super(LightScriptCellReader, self).__init__(fmt, default_language)
        self.ext = self.ext or '.py'
        script = _SCRIPT_EXTENSIONS[self.ext]
        self.default_language = default_language or script['language']
        self.comment = script['comment']
        self.ignore_end_marker = True
        self.explicit_end_marker_required = False
        if fmt and 'cell_markers' in fmt and fmt['cell_markers'] != '+,-':
            self.cell_marker_start, self.cell_marker_end = fmt['cell_markers'].split(',', 1)
            self.start_code_re = re.compile('^' + self.comment + r'\s*' + self.cell_marker_start + r'(.*)$')
            self.end_code_re = re.compile('^' + self.comment + r'\s*' + self.cell_marker_end + r'\s*$')
        else:
            self.start_code_re = re.compile('^' + self.comment + r'\s*\+(.*)$')

    def metadata_and_language_from_option_line(self, line):
        if self.start_code_re.match(line):
            # We want to parse inner most regions as cells.
            # Thus, if we find another region start before the end for this region,
            # we will have ignore the metadata that we found here, and move on to the next cell.
            groups = self.start_code_re.match(line).groups()
            self.language, self.metadata = self.options_to_metadata(groups[0])
            self.ignore_end_marker = False
            if self.cell_marker_start:
                self.explicit_end_marker_required = True
        elif self.simple_start_code_re and self.simple_start_code_re.match(line):
            self.metadata = {}
            self.ignore_end_marker = False
        elif self.cell_marker_end and self.end_code_re.match(line):
            self.metadata = None
            self.cell_type = 'code'

    def options_to_metadata(self, options):
        self.cell_metadata_json = self.cell_metadata_json or is_json_metadata(options)
        title, metadata = text_to_metadata(options, allow_title=True)

        # Cell type
        for cell_type in ['markdown', 'raw', 'md']:
            code = '[{}]'.format(cell_type)
            if code in title:
                title = title.replace(code, '').strip()
                metadata['cell_type'] = cell_type
                if cell_type == 'md':
                    metadata['region_name'] = cell_type
                    metadata['cell_type'] = 'markdown'
                break

        # Spyder has sub cells
        cell_depth = 0
        while title.startswith('%'):
            cell_depth += 1
            title = title[1:]

        if cell_depth:
            metadata['cell_depth'] = cell_depth
            title = title.strip()

        if title:
            metadata['title'] = title

        return None, metadata

    def find_cell_end(self, lines):
        """Return position of end of cell marker, and position of first line after cell"""
        if self.metadata is None and not (self.cell_marker_end and self.end_code_re.match(lines[0])) \
                and paragraph_is_fully_commented(lines, self.comment, self.default_language):
            self.cell_type = 'markdown'
            for i, line in enumerate(lines):
                if _BLANK_LINE.match(line):
                    return i, i + 1, False
            return len(lines), len(lines), False

        if self.metadata is None:
            self.end_code_re = None
        elif not self.cell_marker_end:
            end_of_cell = self.metadata.get('endofcell', '-')
            self.end_code_re = re.compile('^' + self.comment + ' ' + end_of_cell + r'\s*$')

        return self.find_region_end(lines)

    def find_region_end(self, lines):
        """Find the end of the region started with start and end markers"""
        if self.metadata and 'cell_type' in self.metadata:
            self.cell_type = self.metadata.pop('cell_type')
        else:
            self.cell_type = 'code'

        parser = StringParser(self.language or self.default_language)
        for i, line in enumerate(lines):
            # skip cell header
            if self.metadata is not None and i == 0:
                continue

            if parser.is_quoted():
                parser.read_line(line)
                continue

            parser.read_line(line)
            # New code region
            # Simple code pattern in LightScripts must be preceded with a blank line
            if self.start_code_re.match(line) or (
                    self.simple_start_code_re and self.simple_start_code_re.match(line) and
                    (self.cell_marker_start or i == 0 or _BLANK_LINE.match(lines[i - 1]))):

                if self.explicit_end_marker_required:
                    # Metadata here was conditioned on finding an explicit end marker
                    # before the next start marker. So we dismiss it.
                    self.metadata = None
                    self.language = None

                if i > 0 and _BLANK_LINE.match(lines[i - 1]):
                    if i > 1 and _BLANK_LINE.match(lines[i - 2]):
                        return i - 2, i, False
                    return i - 1, i, False
                return i, i, False

            if not self.ignore_end_marker and self.end_code_re:
                if self.end_code_re.match(line):
                    return i, i + 1, True
            elif _BLANK_LINE.match(line):
                if not next_code_is_indented(lines[i:]):
                    if i > 0:
                        return i, i + 1, False
                    if len(lines) > 1 and not _BLANK_LINE.match(lines[1]):
                        return 1, 1, False
                    return 1, 2, False

        return len(lines), len(lines), False


class DoublePercentScriptCellReader(LightScriptCellReader):
    """Read notebook cells from Spyder/VScode scripts (#59)"""
    default_comment_magics = True

    def __init__(self, fmt, default_language=None):
        LightScriptCellReader.__init__(self, fmt, default_language)
        script = _SCRIPT_EXTENSIONS[self.ext]
        self.default_language = default_language or script['language']
        self.comment = script['comment']
        self.start_code_re = re.compile(r"^{}\s*%%(%*)\s(.*)$".format(self.comment))
        self.alternative_start_code_re = re.compile(r"^{}\s*(%%|<codecell>|In\[[0-9 ]*\]:?)\s*$".format(self.comment))
        self.explicit_soc = True

    def metadata_and_language_from_option_line(self, line):
        """Parse code options on the given line. When a start of a code cell
        is found, self.metadata is set to a dictionary."""
        if self.start_code_re.match(line):
            self.language, self.metadata = self.options_to_metadata(line[line.find('%%') + 2:])
        else:
            self.metadata = {}

    def find_cell_content(self, lines):
        """Parse cell till its end and set content, lines_to_next_cell.
        Return the position of next cell start"""
        cell_end_marker, next_cell_start, explicit_eoc = self.find_cell_end(lines)

        # Metadata to dict
        if self.start_code_re.match(lines[0]) or self.alternative_start_code_re.match(lines[0]):
            cell_start = 1
        else:
            cell_start = 0

        # Cell content
        source = lines[cell_start:cell_end_marker]
        self.org_content = copy(source)
        self.content = self.extract_content(source)

        self.lines_to_next_cell = count_lines_to_next_cell(
            cell_end_marker,
            next_cell_start,
            len(lines),
            explicit_eoc)

        return next_cell_start

    def find_cell_end(self, lines):
        """Return position of end of cell marker, and position
        of first line after cell"""

        if self.metadata and 'cell_type' in self.metadata:
            self.cell_type = self.metadata.pop('cell_type')
        else:
            self.cell_type = 'code'

        next_cell = len(lines)
        parser = StringParser(self.language or self.default_language)
        for i, line in enumerate(lines):
            if parser.is_quoted():
                parser.read_line(line)
                continue

            parser.read_line(line)
            if i > 0 and (self.start_code_re.match(line) or self.alternative_start_code_re.match(line)):
                next_cell = i
                break

        if last_two_lines_blank(lines[:next_cell]):
            return next_cell - 2, next_cell, False
        if next_cell > 0 and _BLANK_LINE.match(lines[next_cell - 1]):
            return next_cell - 1, next_cell, False
        return next_cell, next_cell, False


class HydrogenCellReader(DoublePercentScriptCellReader):
    """Read notebook cells from Hydrogen scripts (#59)"""
    default_comment_magics = False


class SphinxGalleryScriptCellReader(ScriptCellReader):  # pylint: disable=W0223
    """Read notebook cells from Sphinx Gallery scripts (#80)"""

    comment = '#'
    default_language = 'python'
    default_comment_magics = True
    twenty_hash = re.compile(r'^#( |)#{19,}\s*$')
    default_markdown_cell_marker = '#' * 79
    markdown_marker = None

    def __init__(self, fmt=None, default_language='python'):
        super(SphinxGalleryScriptCellReader, self).__init__(fmt, default_language)
        self.ext = '.py'
        self.rst2md = (fmt or {}).get('rst2md', False)

    def start_of_new_markdown_cell(self, line):
        """Does this line starts a new markdown cell?
        Then, return the cell marker"""
        for empty_markdown_cell in ['""', "''"]:
            if line == empty_markdown_cell:
                return empty_markdown_cell

        for triple_quote in ['"""', "'''"]:
            if line.startswith(triple_quote):
                return triple_quote

        if self.twenty_hash.match(line):
            return line

        return None

    def metadata_and_language_from_option_line(self, line):
        self.markdown_marker = self.start_of_new_markdown_cell(line)
        if self.markdown_marker:
            self.cell_type = 'markdown'
            if self.markdown_marker != self.default_markdown_cell_marker:
                self.metadata = {'cell_marker': self.markdown_marker}
        else:
            self.cell_type = 'code'

    def find_cell_end(self, lines):
        """Return position of end of cell, and position
        of first line after cell, and whether there was an
        explicit end of cell marker"""

        if self.cell_type == 'markdown':
            # Empty cell "" or ''
            if len(self.markdown_marker) <= 2:
                if len(lines) == 1 or _BLANK_LINE.match(lines[1]):
                    return 0, 2, True
                return 0, 1, True

            # Multi-line comment with triple quote
            if len(self.markdown_marker) == 3:
                for i, line in enumerate(lines):
                    if (i > 0 or line.strip() != self.markdown_marker) and line.rstrip().endswith(self.markdown_marker):
                        explicit_end_of_cell_marker = line.strip() == self.markdown_marker
                        if explicit_end_of_cell_marker:
                            end_of_cell = i
                        else:
                            end_of_cell = i + 1
                        if len(lines) <= i + 1 or _BLANK_LINE.match(
                                lines[i + 1]):
                            return end_of_cell, i + 2, explicit_end_of_cell_marker
                        return end_of_cell, i + 1, explicit_end_of_cell_marker
            else:
                # 20 # or more
                for i, line in enumerate(lines[1:], 1):
                    if not line.startswith(self.comment):
                        if _BLANK_LINE.match(line):
                            return i, i + 1, False
                        return i, i, False

        elif self.cell_type == 'code':
            parser = StringParser('python')
            for i, line in enumerate(lines):
                if parser.is_quoted():
                    parser.read_line(line)
                    continue

                if self.start_of_new_markdown_cell(line):
                    if i > 0 and _BLANK_LINE.match(lines[i - 1]):
                        return i - 1, i, False
                    return i, i, False
                parser.read_line(line)

        return len(lines), len(lines), False

    def find_cell_content(self, lines):
        """Parse cell till its end and set content, lines_to_next_cell.
        Return the position of next cell start"""
        cell_end_marker, next_cell_start, explicit_eoc = self.find_cell_end(lines)

        # Metadata to dict
        cell_start = 0
        if self.cell_type == 'markdown':
            if self.markdown_marker in ['"""', "'''"]:
                # Remove the triple quotes
                if lines[0].strip() == self.markdown_marker:
                    cell_start = 1
                else:
                    lines[0] = lines[0][3:]
                if not explicit_eoc:
                    last = lines[cell_end_marker - 1]
                    lines[cell_end_marker - 1] = last[:last.rfind(self.markdown_marker)]
            if self.twenty_hash.match(self.markdown_marker):
                cell_start = 1
        else:
            self.metadata = {}

        # Cell content
        source = lines[cell_start:cell_end_marker]
        self.org_content = copy(source)

        if self.cell_type == 'code' and self.comment_magics:
            uncomment_magic(source, self.language or self.default_language)

        if self.cell_type == 'markdown' and source:
            if self.markdown_marker.startswith(self.comment):
                source = uncomment(source, self.comment)
            if self.rst2md:
                if rst2md:
                    source = rst2md('\n'.join(source)).splitlines()
                else:
                    raise ImportError('Could not import rst2md from sphinx_gallery.notebook')

        self.content = source

        self.lines_to_next_cell = count_lines_to_next_cell(
            cell_end_marker,
            next_cell_start,
            len(lines),
            explicit_eoc)

        return next_cell_start



import pytest
from jupytext.compare import compare
from jupytext import read, reads, writes
from jupytext.pep8 import next_instruction_is_function_or_class, cell_ends_with_function_or_class
from jupytext.pep8 import cell_ends_with_code, cell_has_code, pep8_lines_between_cells
from .utils import list_notebooks


def test_next_instruction_is_function_or_class():
    text = """@pytest.mark.parametrize('py_file',
    [py_file for py_file in list_notebooks('../jupytext') + list_notebooks('.') if
                                     py_file.endswith('.py')])
def test_no_metadata_when_py_is_pep8(py_file):
    pass
"""
    assert next_instruction_is_function_or_class(text.splitlines())


def test_cell_ends_with_code():
    assert not cell_ends_with_code([])


def test_cell_ends_with_function_or_class():
    text = """class A:
    __init__():
    '''A docstring
with two lines or more'''
        self.a = 0
"""
    assert cell_ends_with_function_or_class(text.splitlines())

    lines = ['#', '#']
    assert not cell_ends_with_function_or_class(lines)

    text = """# two blank line after this class
class A:
    pass


# so we do not need to insert two blank lines below this cell
    """
    assert not cell_ends_with_function_or_class(text.splitlines())

    text = """# All lines
# are commented"""
    assert not cell_ends_with_function_or_class(text.splitlines())

    text = """# Two blank lines after function
def f(x):
    return x


# And a comment here"""
    assert not cell_ends_with_function_or_class(text.splitlines())

    assert not cell_ends_with_function_or_class(['', '#'])


def test_pep8_lines_between_cells():
    prev_lines = """a = a_long_instruction(
    over_two_lines=True)""".splitlines()

    next_lines = """def f(x):
    return x""".splitlines()

    assert cell_ends_with_code(prev_lines)
    assert next_instruction_is_function_or_class(next_lines)
    assert pep8_lines_between_cells(prev_lines, next_lines, '.py') == 2


def test_pep8_lines_between_cells_bis():
    prev_lines = """def f(x):
    return x""".splitlines()

    next_lines = """# A markdown cell

# An instruction
a = 5
""".splitlines()

    assert cell_ends_with_function_or_class(prev_lines)
    assert cell_has_code(next_lines)
    assert pep8_lines_between_cells(prev_lines, next_lines, '.py') == 2

    next_lines = """# A markdown cell

# Only markdown here
# And here
""".splitlines()

    assert cell_ends_with_function_or_class(prev_lines)
    assert not cell_has_code(next_lines)
    assert pep8_lines_between_cells(prev_lines, next_lines, '.py') == 1


def test_pep8():
    text = """import os

path = os.path


# code cell #1, with a comment on f
def f(x):
    return x + 1


# markdown cell #1

# code cell #2 - an instruction
a = 4


# markdown cell #2

# code cell #3 with a comment on g
def g(x):
    return x + 1


# markdown cell #3

# the two lines are:
# - right below the function/class
# - below the last python paragraph (i.e. NOT ABOVE g)

# code cell #4
x = 4
"""
    nb = reads(text, 'py')
    for cell in nb.cells:
        assert not cell.metadata

    text2 = writes(nb, 'py')
    compare(text2, text)


def test_pep8_bis():
    text = """# This is a markdown cell

# a code cell
def f(x):
    return x + 1

# And another markdown cell
# Separated from f by just one line
# As there is no code here
"""
    nb = reads(text, 'py')
    for cell in nb.cells:
        assert not cell.metadata

    text2 = writes(nb, 'py')
    compare(text2, text)


@pytest.mark.parametrize('py_file', [py_file for py_file in list_notebooks('../jupytext') + list_notebooks('.') if
                                     py_file.endswith('.py') and 'folding_markers' not in py_file])
def test_no_metadata_when_py_is_pep8(py_file):
    """This test assumes that all Python files in the jupytext folder follow PEP8 rules"""
    nb = read(py_file)

    for i, cell in enumerate(nb.cells):
        if 'title' in cell.metadata:
            cell.metadata.pop('title')
        if i == 0 and not cell.source:
            assert cell.metadata == {'lines_to_next_cell': 0}, py_file
        else:
            assert not cell.metadata, (py_file, cell.source)



# ---
# jupyter:
#   jupytext:
#     cell_markers: region,endregion
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# region
# %load_ext rpy2.ipython
import pandas as pd

df = pd.DataFrame(
    {
        "Letter": ["a", "a", "a", "b", "b", "b", "c", "c", "c"],
        "X": [4, 3, 5, 2, 1, 7, 7, 5, 9],
        "Y": [0, 4, 3, 6, 7, 10, 11, 9, 13],
        "Z": [1, 2, 3, 1, 2, 3, 1, 2, 3],
    }
)
# endregion

# region magic_args="-i df" language="R"
# library("ggplot2")
# ggplot(data = df) + geom_point(aes(x = X, y = Y, color = Letter, size = Z))
# endregion



# ---
# jupyter:
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

"""
# Jupyter notebook

This notebook is a simple jupyter notebook. It only has markdown and code cells. And it does not contain consecutive markdown cells. We start with an addition:
"""

a = 1
b = 2
a + b

###############################################################################
# Now we return a few tuples

a, b

""
a, b, a+b

###############################################################################
# And this is already the end of the notebook



import re
from nbformat.v4.nbbase import new_notebook, new_markdown_cell, new_code_cell
from jupytext.compare import compare, compare_notebooks
import jupytext
from .utils import skip_if_dict_is_not_ordered


@skip_if_dict_is_not_ordered
def test_read_mostly_py_rmd_file(rmd="""---
title: Simple file
---

```{python, echo=TRUE}
import numpy as np
x = np.arange(0, 2*math.pi, eps)
```

```{python, echo=TRUE}
x = np.arange(0,1,eps)
y = np.abs(x)-.5
```

```{r}
ls()
```

```{r, results="asis", magic_args="-i x"}
cat(stringi::stri_rand_lipsum(3), sep='\n\n')
```
"""):
    nb = jupytext.reads(rmd, 'Rmd')
    compare(nb.cells, [{'cell_type': 'raw',
                        'source': '---\ntitle: Simple file\n---',
                        'metadata': {}},
                       {'cell_type': 'code',
                        'metadata': {'echo': True},
                        'execution_count': None,
                        'source': 'import numpy as np\n'
                                  'x = np.arange(0, 2*math.pi, eps)',
                        'outputs': []},
                       {'cell_type': 'code',
                        'metadata': {'echo': True},
                        'execution_count': None,
                        'source': 'x = np.arange(0,1,eps)\ny = np.abs(x)-.5',
                        'outputs': []},
                       {'cell_type': 'code',
                        'metadata': {},
                        'execution_count': None,
                        'source': '%%R\nls()',
                        'outputs': []},
                       {'cell_type': 'code',
                        'metadata': {'results': 'asis'},
                        'execution_count': None,
                        'source': "%%R -i x\ncat(stringi::"
                                  "stri_rand_lipsum(3), sep='\n\n')",
                        'outputs': []}])

    rmd2 = jupytext.writes(nb, 'Rmd')
    rmd2 = re.sub(r'```{r ', '```{r, ', rmd2)
    rmd2 = re.sub(r'```{python ', '```{python, ', rmd2)
    compare(rmd2, rmd)


def test_markdown_cell_with_code_works(nb=new_notebook(cells=[
    new_markdown_cell("""```python
1 + 1
```""")])):
    text = jupytext.writes(nb, 'Rmd')
    nb2 = jupytext.reads(text, 'Rmd')
    compare_notebooks(nb2, nb)


def test_two_markdown_cell_with_code_works(nb=new_notebook(cells=[
    new_markdown_cell("""```python
1 + 1
```"""),
    new_markdown_cell("""```python
2 + 2
```""")
])):
    text = jupytext.writes(nb, 'Rmd')
    nb2 = jupytext.reads(text, 'Rmd')
    compare_notebooks(nb2, nb)


def test_tags_in_rmd(rmd='''---
jupyter:
  jupytext:
    text_representation:
      extension: .Rmd
      format_name: rmarkdown
      format_version: '1.1'
      jupytext_version: 1.2.3
---

```{python tags=c("parameters")}
p = 1
```
''', nb=new_notebook(cells=[new_code_cell('p = 1', metadata={'tags': ['parameters']})])):
    nb2 = jupytext.reads(rmd, 'Rmd')
    compare_notebooks(nb2, nb)


def round_trip_cell_metadata(cell_metadata):
    nb = new_notebook(metadata={'jupytext': {'main_language': 'python'}},
                      cells=[new_code_cell('1 + 1', metadata=cell_metadata)])
    text = jupytext.writes(nb, 'Rmd')
    nb2 = jupytext.reads(text, 'Rmd')
    compare_notebooks(nb2, nb)


def test_comma_in_metadata(cell_metadata={'a': 'b, c'}):
    round_trip_cell_metadata(cell_metadata)


def test_dict_in_metadata(cell_metadata={'a': {'b': 'c'}}):
    round_trip_cell_metadata(cell_metadata)


def test_list_in_metadata(cell_metadata={'d': ['e']}):
    round_trip_cell_metadata(cell_metadata)



import pytest
from jupytext.compare import compare
from nbformat.v4.nbbase import new_notebook, new_code_cell, new_markdown_cell
import jupytext
from jupytext.compare import compare_notebooks


@pytest.mark.parametrize('blank_lines', range(1, 6))
def test_file_with_blank_lines(blank_lines):
    py_script = """# Markdown cell{0}
# Another one{0}
""".format('\n'.join([''] * blank_lines))

    notebook = jupytext.reads(py_script, 'py')
    py_script2 = jupytext.writes(notebook, 'py')
    compare(py_script2, py_script)


@pytest.mark.parametrize('blank_cells', range(1, 3))
def test_notebook_with_empty_cells(blank_cells):
    notebook = new_notebook(cells=[new_markdown_cell('markdown cell one')] +
                                  [new_code_cell('')] * blank_cells +
                                  [new_markdown_cell('markdown cell two')] +
                                  [new_code_cell('')] * blank_cells,
                            metadata={'jupytext': {'main_language': 'python'}})

    script = jupytext.writes(notebook, 'py')
    notebook2 = jupytext.reads(script, 'py')

    compare_notebooks(notebook2, notebook)



# coding: utf-8
import pytest
import jupytext
from .utils import list_notebooks

try:
    unicode  # Python 2
except NameError:
    unicode = str  # Python 3


@pytest.mark.parametrize('nb_file', list_notebooks() + list_notebooks('Rmd'))
def test_notebook_contents_is_unicode(nb_file):
    nb = jupytext.read(nb_file)

    for cell in nb.cells:
        assert cell.source == '' or isinstance(cell.source, unicode)


def test_write_non_ascii(tmpdir):
    nb = jupytext.reads(u'Non-ascii contênt', 'Rmd')
    jupytext.write(nb, str(tmpdir.join('notebook.Rmd')))
    jupytext.write(nb, str(tmpdir.join('notebook.ipynb')))



from os import path
from io import open
import re
from setuptools import setup, find_packages

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(path.join(this_directory, 'jupytext/version.py')) as f:
    version_file = f.read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    version = version_match.group(1)

setup(
    name='jupytext',
    version=version,
    author='Marc Wouts',
    author_email='marc.wouts@gmail.com',
    description='Jupyter notebooks as Markdown documents, '
                'Julia, Python or R scripts',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/mwouts/jupytext',
    packages=find_packages(exclude=['tests']),
    package_data={'jupytext': ['nbextension/*.*']},
    data_files=[('etc/jupyter/nbconfig/notebook.d', ['jupyter-config/nbconfig/notebook.d/jupytext.json']),
                ('etc/jupyter/jupyter_notebook_config.d', ['jupyter-config/jupyter_notebook_config.d/jupytext.json']),
                ('etc/jupyter/jupyter_server_config.d', ['jupyter-config/jupyter_server_config.d/jupytext.json']),
                ('share/jupyter/nbextensions/jupytext', ['jupytext/nbextension/index.js',
                                                         'jupytext/nbextension/README.md',
                                                         'jupytext/nbextension/jupytext_menu.png',
                                                         'jupytext/nbextension/jupytext_menu_zoom.png',
                                                         'jupytext/nbextension/jupytext.yml']),
                ('share/jupyter/lab/extensions', ['packages/labextension/jupyterlab-jupytext-1.1.0.tgz'])],
    entry_points={'console_scripts': ['jupytext = jupytext.cli:jupytext']},
    tests_require=['pytest'],
    install_requires=['nbformat>=4.0.0', 'pyyaml', 'mock;python_version<"3"'],
    license='MIT',
    classifiers=['Development Status :: 5 - Production/Stable',
                 'License :: OSI Approved :: MIT License',
                 'Environment :: Console',
                 'Framework :: Jupyter',
                 'Intended Audience :: Developers',
                 'Intended Audience :: Science/Research',
                 'Topic :: Text Processing :: Markup',
                 'Programming Language :: Python',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: 3.4',
                 'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: 3.7',
                 'Programming Language :: Python :: 3.8']
)



# -*- coding: utf-8 -*-

from jupytext.compare import compare
import jupytext


def test_read_simple_file(script='''# -*- coding: utf-8 -*-
"""
This is a markdown cell
"""

1 + 2 + 3 + 4
5
6

""
7

#################################
# Another markdown cell

def f(x):
   """Sample docstring"""
   return 4
'''):
    nb = jupytext.reads(script, 'py:sphinx')
    assert nb.cells[0].cell_type == 'code'
    assert nb.cells[0].source == '%matplotlib inline'

    assert nb.cells[1].cell_type == 'markdown'
    assert nb.cells[1].source == 'This is a markdown cell'

    assert nb.cells[2].cell_type == 'code'
    compare(nb.cells[2].source, '''1 + 2 + 3 + 4
5
6''')

    assert nb.cells[3].cell_type == 'code'
    assert nb.cells[3].source == '7'

    assert nb.cells[4].cell_type == 'markdown'
    assert nb.cells[4].source == 'Another markdown cell'

    assert nb.cells[5].cell_type == 'code'
    assert nb.cells[5].source == '''def f(x):
   """Sample docstring"""
   return 4'''
    assert len(nb.cells) == 6

    script2 = jupytext.writes(nb, 'py:sphinx')
    compare(script2, script)


def test_read_more_complex_file(script="""'''This is a markdown cell'''

1 + 2 + 3 + 4
5
6

'''
Another markdown cell'''
#################################
# A third one

'''
Another markdown cell'''
1 + 2 + 3 + 4

# ################################
# A fifth one

'''And a last one
'''
"""):
    nb = jupytext.reads(script, 'py:sphinx')

    assert nb.cells[0].cell_type == 'code'
    assert nb.cells[0].source == '%matplotlib inline'

    assert nb.cells[1].cell_type == 'markdown'
    assert nb.cells[1].source == 'This is a markdown cell'

    assert nb.cells[2].cell_type == 'code'
    compare(nb.cells[2].source, '''1 + 2 + 3 + 4
5
6''')
    assert nb.cells[3].cell_type == 'markdown'
    assert nb.cells[3].source == 'Another markdown cell'

    assert nb.cells[4].cell_type == 'markdown'
    assert nb.cells[4].source == 'A third one'

    assert nb.cells[5].cell_type == 'markdown'
    assert nb.cells[5].source == 'Another markdown cell'

    assert nb.cells[6].cell_type == 'code'
    assert nb.cells[6].source == '1 + 2 + 3 + 4'

    assert nb.cells[7].cell_type == 'markdown'
    assert nb.cells[7].source == 'A fifth one'

    assert nb.cells[8].cell_type == 'markdown'
    assert nb.cells[8].source == 'And a last one'

    assert len(nb.cells) == 9


def test_read_empty_code_cell(script='''"""
Markdown cell
"""


'''):
    nb = jupytext.reads(script, 'py:sphinx')

    assert nb.cells[0].cell_type == 'code'
    assert nb.cells[0].source == '%matplotlib inline'

    assert nb.cells[1].cell_type == 'markdown'
    assert nb.cells[1].source == 'Markdown cell'

    assert nb.cells[2].cell_type == 'code'
    assert nb.cells[2].source == ''

    assert len(nb.cells) == 3

    assert jupytext.writes(nb, 'py:sphinx') == script



# -*- coding: utf-8 -*-

from jupytext.compare import compare
import jupytext


def test_read_simple_file(script=""";; ---
;; title: Simple file
;; ---

;; Here we have some text
;; And below we have some code

(define a 35)
"""):
    for file_extension in ('ss', 'scm'):
        nb = jupytext.reads(script, file_extension)
        assert len(nb.cells) == 3
        assert nb.cells[0].cell_type == 'raw'
        assert nb.cells[0].source == '---\ntitle: Simple file\n---'
        assert nb.cells[1].cell_type == 'markdown'
        assert nb.cells[1].source == 'Here we have some text\n' \
                                     'And below we have some code'
        assert nb.cells[2].cell_type == 'code'
        compare(nb.cells[2].source, '(define a 35)')

        script2 = jupytext.writes(nb, file_extension)
        compare(script2, script)



# -*- coding: utf-8 -*-

from jupytext.compare import compare
import jupytext


def test_read_simple_file(script=""";; ---
;; title: Simple file
;; ---

;; Here we have some text
;; And below we have some code

((fn
  []
  (println "Hello World")))
"""):
    nb = jupytext.reads(script, 'clj')
    assert len(nb.cells) == 3
    assert nb.cells[0].cell_type == 'raw'
    assert nb.cells[0].source == '---\ntitle: Simple file\n---'
    assert nb.cells[1].cell_type == 'markdown'
    assert nb.cells[1].source == 'Here we have some text\n' \
                                 'And below we have some code'
    assert nb.cells[2].cell_type == 'code'
    compare(nb.cells[2].source, """((fn
  []
  (println "Hello World")))""")

    script2 = jupytext.writes(nb, 'clj')
    compare(script2, script)



import pytest
import jupytext
from jupytext.compare import compare, compare_notebooks
from .utils import skip_if_dict_is_not_ordered


@skip_if_dict_is_not_ordered
@pytest.mark.parametrize('md,rmd', [('tags=["remove_cell"]', 'include=FALSE'),
                                    ('tags=["remove_output"]', "results='hide'"),
                                    ('tags=["remove_output"]', 'results="hide"'),
                                    ('tags=["remove_input"]', 'echo=FALSE')])
def test_jupyter_book_options_to_rmarkdown(md, rmd):
    """By default, Jupyter Book tags are mapped to R Markdown options, and vice versa #337"""
    md = '```python ' + md + """
1 + 1
```
"""

    rmd = '```{python ' + rmd + """}
1 + 1
```
"""

    nb_md = jupytext.reads(md, 'md')
    nb_rmd = jupytext.reads(rmd, 'Rmd')
    compare_notebooks(nb_rmd, nb_md)

    md2 = jupytext.writes(nb_rmd, 'md')
    compare(md2, md)

    rmd = rmd.replace('"hide"', "'hide'")
    rmd2 = jupytext.writes(nb_md, 'Rmd')
    compare(rmd2, rmd)


@skip_if_dict_is_not_ordered
@pytest.mark.parametrize('md,rmd', [('hide_input=true hide_output=true', 'include=FALSE'),
                                    ('hide_output=true', "results='hide'"),
                                    ('hide_output=true', 'results="hide"'),
                                    ('hide_input=true', 'echo=FALSE')])
def test_runtools_options_to_rmarkdown(md, rmd):
    """Options set by the runtools extension are mapped to the corresponding R Markdown options
    https://jupyter-contrib-nbextensions.readthedocs.io/en/latest/nbextensions/runtools/readme.html"""
    md = '```python ' + md + """
1 + 1
```
"""

    rmd = '```{python ' + rmd + """}
1 + 1
```
"""

    nb_md = jupytext.reads(md, 'md')
    nb_rmd = jupytext.reads(rmd, fmt={'extension': '.Rmd', 'use_runtools': True})
    compare_notebooks(nb_rmd, nb_md)

    md2 = jupytext.writes(nb_rmd, 'md')
    compare(md2, md)

    rmd = rmd.replace('"hide"', "'hide'")
    rmd2 = jupytext.writes(nb_md, 'Rmd')
    compare(rmd2, rmd)



from nbformat.v4.nbbase import new_code_cell, new_raw_cell, new_markdown_cell, new_notebook
from jupytext.compare import compare, compare_notebooks
import jupytext
from jupytext.combine import combine_inputs_with_outputs


def test_read_mostly_py_markdown_file(markdown="""---
title: Simple file
---

```python
import numpy as np
x = np.arange(0, 2*math.pi, eps)
```

```python
x = np.arange(0,1,eps)
y = np.abs(x)-.5
```

This is
a Markdown cell

```
# followed by a code cell with no language info
```

```
# another code cell


# with two blank lines
```

And the same markdown cell continues

<!-- #raw -->
this is a raw cell
<!-- #endraw -->

```R
ls()
```

```R
cat(stringi::stri_rand_lipsum(3), sep='\n\n')
```
"""):
    nb = jupytext.reads(markdown, 'md')
    assert nb.metadata['jupytext']['main_language'] == 'python'
    compare(nb.cells, [{'cell_type': 'raw',
                        'source': '---\ntitle: Simple file\n---',
                        'metadata': {}},
                       {'cell_type': 'code',
                        'metadata': {},
                        'execution_count': None,
                        'source': 'import numpy as np\n'
                                  'x = np.arange(0, 2*math.pi, eps)',
                        'outputs': []},
                       {'cell_type': 'code',
                        'metadata': {},
                        'execution_count': None,
                        'source': 'x = np.arange(0,1,eps)\ny = np.abs(x)-.5',
                        'outputs': []},
                       {'cell_type': 'markdown',
                        'metadata': {},
                        'source': '''This is
a Markdown cell

```
# followed by a code cell with no language info
```

```
# another code cell


# with two blank lines
```

And the same markdown cell continues'''},
                       {'cell_type': 'raw',
                        'metadata': {},
                        'source': 'this is a raw cell'},
                       {'cell_type': 'code',
                        'metadata': {},
                        'execution_count': None,
                        'source': '%%R\nls()',
                        'outputs': []},
                       {'cell_type': 'code',
                        'metadata': {},
                        'execution_count': None,
                        'source': "%%R\ncat(stringi::"
                                  "stri_rand_lipsum(3), sep='\n\n')",
                        'outputs': []}])

    markdown2 = jupytext.writes(nb, 'md')
    compare(markdown2, markdown)


def test_read_md_and_markdown_regions(markdown="""Some text

<!-- #md -->
A


long
cell
<!-- #endmd -->

<!-- #markdown -->
Another


long
cell
<!-- #endmarkdown -->
"""):
    nb = jupytext.reads(markdown, 'md')
    assert nb.metadata['jupytext']['main_language'] == 'python'
    compare(nb.cells, [new_markdown_cell('Some text'),
                       new_markdown_cell("""A


long
cell""", metadata={'region_name': 'md'}),
                       new_markdown_cell("""Another


long
cell""", metadata={'region_name': 'markdown'})])

    markdown2 = jupytext.writes(nb, 'md')
    compare(markdown2, markdown)


def test_read_mostly_R_markdown_file(markdown="""```R
ls()
```

```R
cat(stringi::stri_rand_lipsum(3), sep='\n\n')
```
"""):
    nb = jupytext.reads(markdown, 'md')
    assert nb.metadata['jupytext']['main_language'] == 'R'
    compare(nb.cells, [{'cell_type': 'code',
                        'metadata': {},
                        'execution_count': None,
                        'source': 'ls()',
                        'outputs': []},
                       {'cell_type': 'code',
                        'metadata': {},
                        'execution_count': None,
                        'source': "cat(stringi::stri_rand_lipsum(3), sep='\n\n')",
                        'outputs': []}])

    markdown2 = jupytext.writes(nb, 'md')
    compare(markdown2, markdown)


def test_read_markdown_file_no_language(markdown="""```
ls
```

```
echo 'Hello World'
```
"""):
    nb = jupytext.reads(markdown, 'md')
    markdown2 = jupytext.writes(nb, 'md')
    compare(markdown2, markdown)


def test_read_julia_notebook(markdown="""```julia
1 + 1
```
"""):
    nb = jupytext.reads(markdown, 'md')
    assert nb.metadata['jupytext']['main_language'] == 'julia'
    assert len(nb.cells) == 1
    assert nb.cells[0].cell_type == 'code'
    markdown2 = jupytext.writes(nb, 'md')
    compare(markdown2, markdown)


def test_split_on_header(markdown="""A paragraph

# H1 Header

## H2 Header

Another paragraph
"""):
    fmt = {'extension': '.md', 'split_at_heading': True}
    nb = jupytext.reads(markdown, fmt)
    assert nb.cells[0].source == 'A paragraph'
    assert nb.cells[1].source == '# H1 Header'
    assert nb.cells[2].source == '## H2 Header\n\nAnother paragraph'
    assert len(nb.cells) == 3
    markdown2 = jupytext.writes(nb, fmt)
    compare(markdown2, markdown)


def test_split_on_header_after_two_blank_lines(markdown="""A paragraph


# H1 Header
"""):
    fmt = {'extension': '.Rmd', 'split_at_heading': True}
    nb = jupytext.reads(markdown, fmt)
    markdown2 = jupytext.writes(nb, fmt)
    compare(markdown2, markdown)


def test_code_cell_with_metadata(markdown="""```python tags=["parameters"]
a = 1
b = 2
```
"""):
    nb = jupytext.reads(markdown, 'md')
    compare(nb.cells[0], new_code_cell(source='a = 1\nb = 2', metadata={'tags': ['parameters']}))
    markdown2 = jupytext.writes(nb, 'md')
    compare(markdown2, markdown)


def test_raw_cell_with_metadata_json(markdown="""<!-- #raw {"key": "value"} -->
raw content
<!-- #endraw -->
"""):
    nb = jupytext.reads(markdown, 'md')
    compare(nb.cells[0], new_raw_cell(source='raw content', metadata={'key': 'value'}))
    markdown2 = jupytext.writes(nb, 'md')
    compare(markdown2, markdown)


def test_raw_cell_with_metadata(markdown="""<!-- #raw key="value" -->
raw content
<!-- #endraw -->
"""):
    nb = jupytext.reads(markdown, 'md')
    compare(nb.cells[0], new_raw_cell(source='raw content', metadata={'key': 'value'}))
    markdown2 = jupytext.writes(nb, 'md')
    compare(markdown2, markdown)


def test_read_raw_cell_markdown_version_1_1(markdown="""---
jupyter:
  jupytext:
    text_representation:
      extension: .md
      format_name: markdown
      format_version: '1.1'
      jupytext_version: 1.1.0
---

```key="value"
raw content
```
"""):
    nb = jupytext.reads(markdown, 'md')
    compare(nb.cells[0], new_raw_cell(source='raw content', metadata={'key': 'value'}))
    md2 = jupytext.writes(nb, 'md')
    assert "format_version: '1.1'" not in md2


def test_read_raw_cell_markdown_version_1_1_with_mimetype(header="""---
jupyter:
  jupytext:
    text_representation:
      extension: .md
      format_name: markdown
      format_version: '1.1'
      jupytext_version: 1.1.0-rc0
  kernelspec:
    display_name: Python 3
    language: python
    name: python3
---
""", markdown_11="""```raw_mimetype="text/restructuredtext"
.. meta::
   :description: Topic: Integrated Development Environments, Difficulty: Easy, Category: Tools
   :keywords: python, introduction, IDE, PyCharm, VSCode, Jupyter, recommendation, tools
```
""", markdown_12="""<!-- #raw raw_mimetype="text/restructuredtext" -->
.. meta::
   :description: Topic: Integrated Development Environments, Difficulty: Easy, Category: Tools
   :keywords: python, introduction, IDE, PyCharm, VSCode, Jupyter, recommendation, tools
<!-- #endraw -->
"""):
    nb = jupytext.reads(header + '\n' + markdown_11, 'md')
    compare(nb.cells[0], new_raw_cell(source=""".. meta::
   :description: Topic: Integrated Development Environments, Difficulty: Easy, Category: Tools
   :keywords: python, introduction, IDE, PyCharm, VSCode, Jupyter, recommendation, tools""",
                                      metadata={'raw_mimetype': 'text/restructuredtext'}))
    md2 = jupytext.writes(nb, 'md')
    assert "format_version: '1.1'" not in md2
    nb.metadata['jupytext']['notebook_metadata_filter'] = '-all'
    md2 = jupytext.writes(nb, 'md')
    compare(md2, markdown_12)


def test_markdown_cell_with_metadata_json(markdown="""<!-- #region {"key": "value"} -->
A long


markdown cell
<!-- #endregion -->
"""):
    nb = jupytext.reads(markdown, 'md')
    compare(nb.cells[0], new_markdown_cell(source='A long\n\n\nmarkdown cell',
                                           metadata={'key': 'value'}))
    markdown2 = jupytext.writes(nb, 'md')
    compare(markdown2, markdown)


def test_markdown_cell_with_metadata(markdown="""<!-- #region key="value" -->
A long


markdown cell
<!-- #endregion -->
"""):
    nb = jupytext.reads(markdown, 'md')
    compare(nb.cells[0], new_markdown_cell(source='A long\n\n\nmarkdown cell',
                                           metadata={'key': 'value'}))
    markdown2 = jupytext.writes(nb, 'md')
    compare(markdown2, markdown)


def test_two_markdown_cells(markdown="""# A header

<!-- #region -->
A long


markdown cell
<!-- #endregion -->
"""):
    nb = jupytext.reads(markdown, 'md')
    compare(nb.cells[0], new_markdown_cell(source='# A header'))
    compare(nb.cells[1], new_markdown_cell(source='A long\n\n\nmarkdown cell'))
    markdown2 = jupytext.writes(nb, 'md')
    compare(markdown2, markdown)


def test_combine_md_version_one():
    markdown = """---
jupyter:
  jupytext:
    text_representation:
      extension: .md
      format_name: markdown
      format_version: '1.0'
      jupytext_version: 1.0.0
  kernelspec:
    display_name: Python 3
    language: python
    name: python3
---

A short markdown cell

```
a raw cell
```

```python
1 + 1
```
"""
    # notebook read from markdown file in version 1.0
    nb_source = jupytext.reads(markdown, 'md')

    # actual notebook has metadata
    nb_meta = jupytext.reads(markdown, 'md')
    for cell in nb_meta.cells:
        cell.metadata = {'key': 'value'}

    combine_inputs_with_outputs(nb_source, nb_meta)
    for cell in nb_source.cells:
        assert cell.metadata == {'key': 'value'}, cell.source


def test_jupyter_cell_is_not_split():
    text = """Here we have a markdown
file with a jupyter code cell

```python
1 + 1


2 + 2
```

the code cell should become a Jupyter cell.
"""
    nb = jupytext.reads(text, 'md')
    assert nb.cells[0].cell_type == 'markdown'
    compare(nb.cells[0].source, """Here we have a markdown
file with a jupyter code cell""")

    assert nb.cells[1].cell_type == 'code'
    compare(nb.cells[1].source, """1 + 1


2 + 2""")

    assert nb.cells[2].cell_type == 'markdown'
    compare(nb.cells[2].source, "the code cell should become a Jupyter cell.")
    assert len(nb.cells) == 3


def test_indented_code_is_not_split():
    text = """Here we have a markdown
file with an indented code cell

    1 + 1


    2 + 2

the code cell should not become a Jupyter cell,
nor be split into two pieces."""
    nb = jupytext.reads(text, 'md')
    compare(nb.cells[0].source, text)
    assert nb.cells[0].cell_type == 'markdown'
    assert len(nb.cells) == 1


def test_non_jupyter_code_is_not_split():
    text = """Here we have a markdown
file with a non-jupyter code cell

```{.python}
1 + 1


2 + 2
```

the code cell should not become a Jupyter cell,
nor be split into two pieces."""
    nb = jupytext.reads(text, 'md')
    compare(nb.cells[0].source, text)
    assert nb.cells[0].cell_type == 'markdown'
    assert len(nb.cells) == 1


def test_read_markdown_idl(no_jupytext_version_number, text='''---
jupyter:
  kernelspec:
    display_name: IDL [conda env:gdl] *
    language: IDL
    name: conda-env-gdl-idl
---

# A sample IDL Markdown notebook

```idl
a = 1
```
'''):
    nb = jupytext.reads(text, 'md')
    assert len(nb.cells) == 2
    assert nb.cells[1].cell_type == 'code'
    assert nb.cells[1].source == 'a = 1'

    text2 = jupytext.writes(nb, 'md')
    compare(text2, text)


def test_read_markdown_IDL(no_jupytext_version_number, text='''---
jupyter:
  kernelspec:
    display_name: IDL [conda env:gdl] *
    language: IDL
    name: conda-env-gdl-idl
---

# A sample IDL Markdown notebook

```IDL
a = 1
```
'''):
    nb = jupytext.reads(text, 'md')
    assert len(nb.cells) == 2
    assert nb.cells[1].cell_type == 'code'
    assert nb.cells[1].source == 'a = 1'

    text2 = jupytext.writes(nb, 'md')
    compare(text2, text.replace('```IDL', '```idl'))


def test_inactive_cell(text='''```python active="md"
# This becomes a raw cell in Jupyter
```
''', expected=new_notebook(
    cells=[new_raw_cell('# This becomes a raw cell in Jupyter',
                        metadata={'active': 'md'})])):
    nb = jupytext.reads(text, 'md')
    compare_notebooks(nb, expected)
    text2 = jupytext.writes(nb, 'md')
    compare(text2, text)


def test_inactive_cell_using_tag(text='''```python tags=["active-md"]
# This becomes a raw cell in Jupyter
```
''', expected=new_notebook(
    cells=[new_raw_cell('# This becomes a raw cell in Jupyter',
                        metadata={'tags': ['active-md']})])):
    nb = jupytext.reads(text, 'md')
    compare_notebooks(nb, expected)
    text2 = jupytext.writes(nb, 'md')
    compare(text2, text)


def test_inactive_cell_using_noeval(text='''This is text

```python .noeval
# This is python code.
# It should not become a code cell
```
'''):
    expected = new_notebook(cells=[new_markdown_cell(text[:-1])])
    nb = jupytext.reads(text, 'md')
    compare_notebooks(nb, expected)
    text2 = jupytext.writes(nb, 'md')
    compare(text2, text)


def test_noeval_followed_by_code_works(text='''```python .noeval
# Not a code cell in Jupyter
```

```python
1 + 1
```
''', expected=new_notebook(cells=[new_markdown_cell('''```python .noeval
# Not a code cell in Jupyter
```'''),
                                  new_code_cell('1 + 1')])):
    nb = jupytext.reads(text, 'md')
    compare_notebooks(nb, expected)
    text2 = jupytext.writes(nb, 'md')
    compare(text2, text)


def test_markdown_cell_with_code_works(nb=new_notebook(cells=[
    new_markdown_cell("""```python
1 + 1
```""")])):
    text = jupytext.writes(nb, 'md')
    nb2 = jupytext.reads(text, 'md')
    compare_notebooks(nb2, nb)


def test_markdown_cell_with_noeval_code_works(nb=new_notebook(cells=[
    new_markdown_cell("""```python .noeval
1 + 1
```""")])):
    text = jupytext.writes(nb, 'md')
    nb2 = jupytext.reads(text, 'md')
    compare_notebooks(nb2, nb)


def test_two_markdown_cell_with_code_works(nb=new_notebook(cells=[
    new_markdown_cell("""```python
1 + 1
```"""),
    new_markdown_cell("""```python
2 + 2
```""")
])):
    text = jupytext.writes(nb, 'md')
    nb2 = jupytext.reads(text, 'md')
    compare_notebooks(nb2, nb)


def test_two_markdown_cell_with_no_language_code_works(nb=new_notebook(cells=[
    new_markdown_cell("""```
1 + 1
```"""),
    new_markdown_cell("""```
2 + 2
```""")
])):
    text = jupytext.writes(nb, 'md')
    nb2 = jupytext.reads(text, 'md')
    compare_notebooks(nb2, nb)


def test_notebook_with_python3_magic(no_jupytext_version_number,
                                     nb=new_notebook(metadata={
                                         'kernelspec': {'display_name': 'Python 3', 'language': 'python',
                                                        'name': 'python3'}}, cells=[
                                         new_code_cell('%%python2\na = 1\nprint a'),
                                         new_code_cell('%%python3\nb = 2\nprint(b)')]),
                                     text="""---
jupyter:
  kernelspec:
    display_name: Python 3
    language: python
    name: python3
---

```python2
a = 1
print a
```

```python3
b = 2
print(b)
```
"""):
    md = jupytext.writes(nb, 'md')
    compare(md, text)

    nb2 = jupytext.reads(md, 'md')
    compare_notebooks(nb2, nb)


def test_update_metadata_filter(
        no_jupytext_version_number,
        org="""---
jupyter:
  kernelspec:
    display_name: Python 3
    language: python
    name: python3
  extra:
    key: value
---
""", target="""---
jupyter:
  extra:
    key: value
  jupytext:
    notebook_metadata_filter: extra
  kernelspec:
    display_name: Python 3
    language: python
    name: python3
---
"""):
    nb = jupytext.reads(org, 'md')
    text = jupytext.writes(nb, 'md')
    compare(text, target)


def test_update_metadata_filter_2(
        no_jupytext_version_number,
        org="""---
jupyter:
  jupytext:
    notebook_metadata_filter: -extra
  kernelspec:
    display_name: Python 3
    language: python
    name: python3
  extra:
    key: value
---
""", target="""---
jupyter:
  jupytext:
    notebook_metadata_filter: -extra
  kernelspec:
    display_name: Python 3
    language: python
    name: python3
---
"""):
    nb = jupytext.reads(org, 'md')
    text = jupytext.writes(nb, 'md')
    compare(text, target)



# -*- coding: utf-8 -*-
from jupytext.compare import compare
from nbformat.v4.nbbase import new_notebook, new_markdown_cell
from jupytext.compare import compare_notebooks
import jupytext
from .utils import requires_pandoc


@requires_pandoc
def test_pandoc_implicit(markdown='''# Lorem ipsum

**Lorem ipsum** dolor sit amet, consectetur adipiscing elit. Nunc luctus
bibendum felis dictum sodales.

``` code
print("hello")
```
'''):
    nb = jupytext.reads(markdown, 'md:pandoc')
    markdown2 = jupytext.writes(nb, 'md')

    nb2 = jupytext.reads(markdown2, 'md')
    compare_notebooks(nb2, nb)

    markdown3 = jupytext.writes(nb2, 'md')
    compare(markdown3, markdown2)


@requires_pandoc
def test_pandoc_explicit(markdown='''::: {.cell .markdown}
# Lorem

**Lorem ipsum** dolor sit amet, consectetur adipiscing elit. Nunc luctus
bibendum felis dictum sodales.
:::'''):
    nb = jupytext.reads(markdown, 'md')

    markdown2 = jupytext.writes(nb, 'md')
    compare('\n'.join(markdown2.splitlines()[12:]), markdown)


@requires_pandoc
def test_pandoc_utf8_in_md(markdown='''::: {.cell .markdown}
# Utf-8 support

This is the greek letter $\\pi$: π
:::'''):
    nb = jupytext.reads(markdown, 'md')

    markdown2 = jupytext.writes(nb, 'md')
    compare('\n'.join(markdown2.splitlines()[12:]), markdown)


@requires_pandoc
def test_pandoc_utf8_in_nb(nb=new_notebook(cells=[new_markdown_cell('''# Utf-8 support

This is the greek letter $\\pi$: π''')])):
    markdown = jupytext.writes(nb, 'md:pandoc')
    nb2 = jupytext.reads(markdown, 'md:pandoc')
    nb2.metadata.pop('jupytext')
    compare_notebooks(nb, nb2, 'md:pandoc')



from nbformat.v4.nbbase import new_notebook, new_code_cell, new_markdown_cell
from jupytext.compare import compare
from jupytext import reads, writes
from jupytext.combine import combine_inputs_with_outputs


def test_write_reload_simple_notebook():
    nb1 = new_notebook(cells=[
        new_markdown_cell('A markdown cell', metadata={'md': 'value'}),
        new_code_cell('1 + 1'),
        new_markdown_cell('A markdown cell', metadata={'md': 'value'}),
        new_code_cell("""def f(x):
    return x""", metadata={'md': 'value'}),
        new_markdown_cell('A markdown cell', metadata={'md': 'value'}),
        new_code_cell("""def g(x):
    return x


def h(x):
    return x
""", metadata={'md': 'value'})])

    text = writes(nb1, 'py:bare')
    nb2 = reads(text, 'py:bare')
    combine_inputs_with_outputs(nb2, nb1, 'py:bare')
    nb2.metadata.pop('jupytext')

    assert len(nb2.cells) == 7
    nb1.cells = nb1.cells[:5]
    nb2.cells = nb2.cells[:5]
    compare(nb2, nb1)



import pytest
from jupytext.compare import compare
from jupytext.cell_metadata import rmd_options_to_metadata, metadata_to_rmd_options, parse_rmd_options
from jupytext.cell_metadata import _IGNORE_CELL_METADATA, RMarkdownOptionParsingError, try_eval_metadata
from jupytext.cell_metadata import text_to_metadata, metadata_to_text
from jupytext.metadata_filter import filter_metadata
from .utils import skip_if_dict_is_not_ordered

SAMPLES = [('r', ('R', {})),
           ('r plot_1, dpi=72, fig.path="fig_path/"',
            ('R', {'name': 'plot_1', 'dpi': 72, 'fig.path': 'fig_path/'})),
           ('r plot_1, bool=TRUE, fig.path="fig_path/"',
            ('R', {'name': 'plot_1', 'bool': True,
                   'fig.path': 'fig_path/'})),
           ('r echo=FALSE',
            ('R', {'tags': ['remove_input']})),
           ('r plot_1, echo=TRUE',
            ('R', {'name': 'plot_1', 'echo': True})),
           ('python echo=if a==5 then TRUE else FALSE',
            ('python', {'echo': '#R_CODE#if a==5 then TRUE else FALSE'})),
           ('python noname, tags=c("a", "b", "c"), echo={sum(a+c(1,2))>1}',
            ('python', {'name': 'noname', 'tags': ['a', 'b', 'c'],
                        'echo': '#R_CODE#{sum(a+c(1,2))>1}'})),
           ('python active="ipynb,py"',
            ('python', {'active': 'ipynb,py'})),
           ('python include=FALSE, active="Rmd"',
            ('python', {'active': 'Rmd', 'tags': ['remove_cell']})),
           ('r chunk_name, include=FALSE, active="Rmd"',
            ('R',
             {'name': 'chunk_name', 'active': 'Rmd', 'tags': ['remove_cell']})),
           ('python tags=c("parameters")',
            ('python', {'tags': ['parameters']}))]


@pytest.mark.parametrize('options,language_and_metadata', SAMPLES)
def test_parse_rmd_options(options, language_and_metadata):
    compare(rmd_options_to_metadata(options), language_and_metadata)


@skip_if_dict_is_not_ordered
@pytest.mark.parametrize('options,language_and_metadata', SAMPLES)
def test_build_options(options, language_and_metadata):
    compare(metadata_to_rmd_options(*language_and_metadata), options)


@pytest.mark.parametrize('options,language_and_metadata', SAMPLES)
def test_build_options_random_order(options, language_and_metadata):
    # Older python has no respect for order...
    # assert to_chunk_options(metadata) == options

    def split_and_strip(opt):
        set([o.strip() for o in opt.split(',')])

    assert (split_and_strip(metadata_to_rmd_options(*language_and_metadata)) ==
            split_and_strip(options))


@pytest.mark.parametrize('options', ['a={)', 'name, name2',
                                     'a=}', 'b=]', 'c=['])
def test_parsing_error(options):
    with pytest.raises(RMarkdownOptionParsingError):
        parse_rmd_options(options)


def test_ignore_metadata():
    metadata = {'trusted': True, 'tags': ['remove_input']}
    metadata = filter_metadata(metadata, None, _IGNORE_CELL_METADATA)
    assert metadata_to_rmd_options('R', metadata) == 'r echo=FALSE'


def test_filter_metadata():
    assert filter_metadata({'scrolled': True}, None, _IGNORE_CELL_METADATA) == {}


def test_try_eval_metadata():
    metadata = {'list': 'list("a",5)',
                'c': 'c(1,2,3)'}
    try_eval_metadata(metadata, 'list')
    try_eval_metadata(metadata, 'c')
    assert metadata == {'list': ['a', 5], 'c': [1, 2, 3]}


pytestmark = skip_if_dict_is_not_ordered


def test_language_no_metadata(text='python', value=('python', {})):
    compare(text_to_metadata(text), value)
    assert metadata_to_text(*value) == text.strip()


def test_only_metadata(text='key="value"', value=('', {'key': 'value'})):
    compare(text_to_metadata(text), value)
    assert metadata_to_text(*value) == text


def test_only_metadata_2(text='key="value"', value=('', {'key': 'value'})):
    compare(text_to_metadata(text, allow_title=True), value)
    assert metadata_to_text(*value) == text


def test_no_language(text='.class', value=('', {'.class': None})):
    compare(text_to_metadata(text), value)
    assert metadata_to_text(*value) == text


def test_language_metadata_no_space(text='python{"a":1}', value=('python', {'a': 1})):
    compare(text_to_metadata(text), value)
    assert metadata_to_text(*value) == 'python a=1'


def test_title_no_metadata(text='title', value=('title', {})):
    compare(text_to_metadata(text, allow_title=True), value)
    assert metadata_to_text(*value) == text.strip()


def test_simple_metadata(text='python string="value" number=1.0 array=["a", "b"]',
                         value=('python', {'string': 'value', 'number': 1.0, 'array': ['a', 'b']})):
    compare(text_to_metadata(text), value)
    assert metadata_to_text(*value) == text


def test_simple_metadata_with_spaces(text='python string = "value" number = 1.0 array = ["a",  "b"]',
                                     value=('python', {'string': 'value', 'number': 1.0, 'array': ['a', 'b']})):
    compare(text_to_metadata(text), value)
    assert metadata_to_text(*value) == 'python string="value" number=1.0 array=["a", "b"]'


def test_title_and_relax_json(text="cell title string='value' number=1.0 array=['a', \"b\"]",
                              value=('cell title', {'string': 'value', 'number': 1.0, 'array': ['a', 'b']})):
    compare(text_to_metadata(text, allow_title=True), value)
    assert metadata_to_text(*value) == 'cell title string="value" number=1.0 array=["a", "b"]'


def test_title_and_json_dict(text='cell title {"string": "value", "number": 1.0, "array": ["a", "b"]}',
                             value=('cell title', {'string': 'value', 'number': 1.0, 'array': ['a', 'b']})):
    compare(text_to_metadata(text, allow_title=True), value)
    assert metadata_to_text(*value) == 'cell title string="value" number=1.0 array=["a", "b"]'


@pytest.mark.parametrize('allow_title', [True, False])
def test_attribute(allow_title):
    text = ".class"
    value = ('', {'.class': None})
    compare(text_to_metadata(text, allow_title), value)
    assert metadata_to_text(*value) == text


def test_language_and_attribute(text="python .class", value=('python', {'.class': None})):
    compare(text_to_metadata(text), value)
    assert metadata_to_text(*value) == text


def test_title_and_attribute(text="This is my title. .class", value=('This is my title.', {'.class': None})):
    compare(text_to_metadata(text, allow_title=True), value)
    assert metadata_to_text(*value) == text


def test_values_with_equal_signs_inside(text='python string="value=5"',
                                        value=('python', {'string': 'value=5'})):
    compare(text_to_metadata(text), value)
    assert metadata_to_text(*value) == text


def test_incorrectly_encoded(text="this is an incorrect expression d={{4 b=3"):
    value = text_to_metadata(text, allow_title=True)
    assert metadata_to_text(*value) == text



import pytest
from nbformat.v4.nbbase import new_notebook
from jupytext import reads, writes
from jupytext.metadata_filter import filter_metadata, metadata_filter_as_dict


def to_dict(keys):
    return {key: None for key in keys}


@pytest.mark.parametrize('metadata_filter_string,metadata_filter_dict',
                         [('all, -widgets,-varInspector',
                           {'additional': 'all', 'excluded': ['widgets', 'varInspector']}),
                          ('toc', {'additional': ['toc']}),
                          ('+ toc', {'additional': ['toc']}),
                          ('preserve,-all', {'additional': ['preserve'], 'excluded': 'all'}),
                          ('ExecuteTime, autoscroll, -hide_output',
                           {'additional': ['ExecuteTime', 'autoscroll'], 'excluded': ['hide_output']})])
def test_string_to_dict_conversion(metadata_filter_string, metadata_filter_dict):
    assert metadata_filter_as_dict(metadata_filter_string) == metadata_filter_dict


def test_metadata_filter_as_dict():
    assert metadata_filter_as_dict(True) == metadata_filter_as_dict('all')
    assert metadata_filter_as_dict(False) == metadata_filter_as_dict('-all')
    assert metadata_filter_as_dict({'excluded': 'all'}) == metadata_filter_as_dict('-all')


def test_metadata_filter_default():
    assert filter_metadata(to_dict(['technical', 'user', 'preserve']), None, '-technical'
                           ) == to_dict(['user', 'preserve'])
    assert filter_metadata(to_dict(['technical', 'user', 'preserve']), None, 'preserve,-all'
                           ) == to_dict(['preserve'])


def test_metadata_filter_user_plus_default():
    assert filter_metadata(to_dict(['technical', 'user', 'preserve']), '-user', '-technical'
                           ) == to_dict(['preserve'])
    assert filter_metadata(to_dict(['technical', 'user', 'preserve']), 'all,-user', '-technical'
                           ) == to_dict(['preserve', 'technical'])
    assert filter_metadata(to_dict(['technical', 'user', 'preserve']), 'user', 'preserve,-all'
                           ) == to_dict(['user', 'preserve'])


def test_metadata_filter_user_overrides_default():
    assert filter_metadata(to_dict(['technical', 'user', 'preserve']), 'all,-user', '-technical'
                           ) == to_dict(['technical', 'preserve'])
    assert filter_metadata(to_dict(['technical', 'user', 'preserve']), 'user,-all', 'preserve'
                           ) == to_dict(['user'])


def test_negative_cell_metadata_filter():
    assert filter_metadata(to_dict(['exectime']), '-linesto', '-exectime') == to_dict([])


def test_cell_metadata_filter_is_updated():
    text = """---
jupyter:
  jupytext:
    cell_metadata_filter: -all
---

```{r cache=FALSE}
1+1
```
"""
    nb = reads(text, 'Rmd')
    assert nb.metadata['jupytext']['cell_metadata_filter'] == 'cache,-all'

    text2 = writes(nb, 'Rmd')
    assert text.splitlines()[-3:] == text2.splitlines()[-3:]


def test_notebook_metadata_all():
    nb = new_notebook(metadata={'user_metadata': [1, 2, 3],
                                'jupytext': {'notebook_metadata_filter': 'all'}})
    text = writes(nb, 'md')
    assert 'user_metadata' in text


def test_notebook_metadata_none():
    nb = new_notebook(metadata={'jupytext': {'notebook_metadata_filter': '-all'}})
    text = writes(nb, 'md')
    assert '---' not in text
